/*****************************************************************************
 * analyse.c: h264 encoder library
 *****************************************************************************
 * Copyright (C) 2003 Laurent Aimar
 * $Id: analyse.c,v 1.4 2004/04/24 10:21:22 titer Exp $
 *
 * Authors: Laurent Aimar <fenrir@via.ecp.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <math.h>

#include "../core/common.h"
#include "../core/macroblock.h"
#include "me.h"

static const uint8_t block_idx_x[16] = { 0, 1, 0, 1, 2, 3, 2, 3, 0, 1, 0, 1, 2, 3, 2, 3 };
static const uint8_t block_idx_y[16] = { 0, 0, 1, 1, 0, 0, 1, 1, 2, 2, 3, 3, 2, 2, 3, 3 };

/*
 * Handle intra mb
 */
/* Max = 4 */
static void predict_16x16_mode_available( unsigned int i_neighbour, int *mode, int *pi_count )
{
    if( ( i_neighbour & (MB_LEFT|MB_TOP) ) == (MB_LEFT|MB_TOP) )
    {
        /* top and left avaible */
        *mode++ = I_PRED_16x16_V;
        *mode++ = I_PRED_16x16_H;
        *mode++ = I_PRED_16x16_DC;
        *mode++ = I_PRED_16x16_P;
        *pi_count = 4;
    }
    else if( ( i_neighbour & MB_LEFT ) )
    {
        /* left available*/
        *mode++ = I_PRED_16x16_DC_LEFT;
        *mode++ = I_PRED_16x16_H;
        *pi_count = 2;
    }
    else if( ( i_neighbour & MB_TOP ) )
    {
        /* top available*/
        *mode++ = I_PRED_16x16_DC_TOP;
        *mode++ = I_PRED_16x16_V;
        *pi_count = 2;
    }
    else
    {
        /* none avaible */
        *mode = I_PRED_16x16_DC_128;
        *pi_count = 1;
    }
}

/* Max = 4 */
static void predict_8x8_mode_available( unsigned int i_neighbour, int *mode, int *pi_count )
{
    if( ( i_neighbour & (MB_LEFT|MB_TOP) ) == (MB_LEFT|MB_TOP) )
    {
        /* top and left avaible */
        *mode++ = I_PRED_CHROMA_V;
        *mode++ = I_PRED_CHROMA_H;
        *mode++ = I_PRED_CHROMA_DC;
        *mode++ = I_PRED_CHROMA_P;
        *pi_count = 4;
    }
    else if( ( i_neighbour & MB_LEFT ) )
    {
        /* left available*/
        *mode++ = I_PRED_CHROMA_DC_LEFT;
        *mode++ = I_PRED_CHROMA_H;
        *pi_count = 2;
    }
    else if( ( i_neighbour & MB_TOP ) )
    {
        /* top available*/
        *mode++ = I_PRED_CHROMA_DC_TOP;
        *mode++ = I_PRED_CHROMA_V;
        *pi_count = 2;
    }
    else
    {
        /* none avaible */
        *mode = I_PRED_CHROMA_DC_128;
        *pi_count = 1;
    }
}

/* MAX = 8 */
static void predict_4x4_mode_available( unsigned int i_neighbour, int idx, int *mode, int *pi_count )
{
    int b_a, b_b, b_c;
    static const unsigned int needmb[16] =
    {
        MB_LEFT|MB_TOP, MB_TOP,
        MB_LEFT,        MB_PRIVATE,
        MB_TOP,         MB_TOP|MB_TOPRIGHT,
        0,              MB_PRIVATE,
        MB_LEFT,        0,
        MB_LEFT,        MB_PRIVATE,
        0,              MB_PRIVATE,
        0,              MB_PRIVATE
    };

    /* FIXME even when b_c == 0 there is some case where missing pixels
     * are emulated and thus more mode are available TODO
     * analysis and encode should be fixed too */
    b_a = (needmb[idx]&i_neighbour&MB_LEFT) == (needmb[idx]&MB_LEFT);
    b_b = (needmb[idx]&i_neighbour&MB_TOP) == (needmb[idx]&MB_TOP);
    b_c = (needmb[idx]&i_neighbour&(MB_TOPRIGHT|MB_PRIVATE)) == (needmb[idx]&(MB_TOPRIGHT|MB_PRIVATE));

    if( b_a && b_b )
    {
        *mode++ = I_PRED_4x4_DC;
        *mode++ = I_PRED_4x4_H;
        *mode++ = I_PRED_4x4_V;
        *mode++ = I_PRED_4x4_DDR;
        *mode++ = I_PRED_4x4_VR;
        *mode++ = I_PRED_4x4_HD;
        *mode++ = I_PRED_4x4_HU;

        *pi_count = 7;

        if( b_c )
        {
            *mode++ = I_PRED_4x4_DDL;
            *mode++ = I_PRED_4x4_VL;
            (*pi_count) += 2;
        }
    }
    else if( b_a && !b_b )
    {
        *mode++ = I_PRED_4x4_DC_LEFT;
        *mode++ = I_PRED_4x4_H;
        *pi_count = 2;
    }
    else if( !b_a && b_b )
    {
        *mode++ = I_PRED_4x4_DC_TOP;
        *mode++ = I_PRED_4x4_V;
        *pi_count = 2;
    }
    else
    {
        *mode++ = I_PRED_4x4_DC_128;
        *pi_count = 1;
    }
}

typedef struct
{
    /* conduct the analysis using this lamda and QP */
    int i_lambda;
    int i_qp;


    /* I: Intra part */
    /* Luma part 16x16 and 4x4 modes stats */
    int i_sad_i16x16;
    int i_predict16x16;

    int i_sad_i4x4;
    int i_predict4x4[4][4];

    /* Chroma part */
    int i_sad_i8x8;
    int i_predict8x8;

    /* II: Inter part P frame */
    int i_sad_p16x16;
    int i_ref_p16x16;
    int i_mv_p16x16[2];

    int i_sad_p16x8;
    int i_ref_p16x8;
    int i_mv_p16x8[2][2];

    int i_sad_p8x16;
    int i_ref_p8x16;
    int i_mv_p8x16[2][2];

    int i_sad_p8x8;
    int i_ref_p8x8;
    int i_sub_partition_p8x8[4];
    int i_mv_p8x8[4][4][2];

    /* II: Inter part B frame */
    int i_sad_b16x16_l0;
    int i_ref_b16x16_l0;
    int i_mv_b16x16_l0[2];

    int i_sad_b16x16_l1;
    int i_ref_b16x16_l1;
    int i_mv_b16x16_l1[2];

    int i_sad_b16x16_bi;    /* used the same ref and mv as l0 and l1 (at least for now) */

    int i_sad_b16x8_l0;
    int i_ref_b16x8_l0;
    int i_mv_b16x8_l0[2];

    int i_sad_b8x16_l0;
    int i_ref_b8x16_l0;
    int i_mv_b8x16_l0[2];



} x264_mb_analysis_t;


static const int i_qp0_cost_table[52] =
{
   1, 1, 1, 1, 1, 1, 1, 1,
   1, 1, 1, 1,
   1, 1, 1, 1, 2, 2, 2, 2,
   3, 3, 3, 4, 4, 4, 5, 6,
   6, 7, 8, 9,10,11,13,14,
  16,18,20,23,25,29,32,36,
  40,45,51,57,64,72,81,91
};

static void x264_mb_analyse_intra_chroma( x264_t *h, x264_mb_analysis_t *res )
{
    int i;

    int i_max;
    int predict_mode[9];

    uint8_t *p_dstc[2], *p_srcc[2];
    int      i_dstc[2], i_srcc[2];

    /* 8x8 prediction selection for chroma */
    p_dstc[0] = h->mb.pic.p_fdec[1]; i_dstc[0] = h->mb.pic.i_fdec[1];
    p_dstc[1] = h->mb.pic.p_fdec[2]; i_dstc[1] = h->mb.pic.i_fdec[2];
    p_srcc[0] = h->mb.pic.p_img[1];  i_srcc[0] = h->mb.pic.i_img[1];
    p_srcc[1] = h->mb.pic.p_img[2];  i_srcc[1] = h->mb.pic.i_img[2];

    predict_8x8_mode_available( h->mb.i_neighbour, predict_mode, &i_max );
    res->i_sad_i8x8 = -1;
    for( i = 0; i < i_max; i++ )
    {
        int i_sad;
        int i_mode;

        i_mode = predict_mode[i];

        /* we do the prediction */
        h->predict_8x8[i_mode]( p_dstc[0], i_dstc[0] );
        h->predict_8x8[i_mode]( p_dstc[1], i_dstc[1] );

        /* we calculate the cost */
        i_sad = h->pixf.satd[PIXEL_8x8]( p_dstc[0], i_dstc[0], p_srcc[0], i_srcc[0] ) +
                h->pixf.satd[PIXEL_8x8]( p_dstc[1], i_dstc[1], p_srcc[1], i_srcc[1] ) +
                res->i_lambda * bs_size_ue( x264_mb_pred_mode8x8_fix[i_mode] );

        /* if i_score is lower it is better */
        if( res->i_sad_i8x8 == -1 || res->i_sad_i8x8 > i_sad )
        {
            res->i_predict8x8 = i_mode;
            res->i_sad_i8x8     = i_sad;
        }
    }
}

static void x264_mb_analyse_intra( x264_t *h, x264_mb_analysis_t *res )
{
    const unsigned int flags = h->sh.i_type == SLICE_TYPE_I ? h->param.analyse.intra : h->param.analyse.inter;
    const int i_src = h->mb.pic.i_img[0];
    uint8_t  *p_src = h->mb.pic.p_img[0];
    const int i_dst = h->mb.pic.i_fdec[0];
    uint8_t  *p_dst = h->mb.pic.p_fdec[0];

    int i, idx;

    int i_max;
    int predict_mode[9];

    /*---------------- Try all mode and calculate their score ---------------*/

    /* 16x16 prediction selection */
    res->i_sad_i16x16 = -1;
    predict_16x16_mode_available( h->mb.i_neighbour, predict_mode, &i_max );
    if( !( flags & X264_ANALYSE_I16x16 ) )
    {
        i_max = 1;
    }

    for( i = 0; i < i_max; i++ )
    {
        int i_sad;
        int i_mode;

        i_mode = predict_mode[i];

        /* we do the prediction */
        h->predict_16x16[i_mode]( p_dst, i_dst );

        /* we calculate the diff and get the square sum of the diff */
        i_sad = h->pixf.satd[PIXEL_16x16]( p_dst, i_dst, p_src, i_src ) +
                res->i_lambda * bs_size_ue( x264_mb_pred_mode16x16_fix[i_mode] );
        /* if i_score is lower it is better */
        if( res->i_sad_i16x16 == -1 || res->i_sad_i16x16 > i_sad )
        {
            res->i_predict16x16 = i_mode;
            res->i_sad_i16x16     = i_sad;
        }
    }

    /* 4x4 prediction selection */
    res->i_sad_i4x4 = -1;
    if( flags & X264_ANALYSE_I4x4 )
    {
        res->i_sad_i4x4 = 0;
        for( idx = 0; idx < 16; idx++ )
        {
            uint8_t *p_src_by;
            uint8_t *p_dst_by;
            int     i_best;
            int x, y;
            int i_pred_mode;

            i_pred_mode= x264_mb_predict_intra4x4_mode( h, idx );
            x = block_idx_x[idx];
            y = block_idx_y[idx];

            p_src_by = p_src + 4 * x + 4 * y * i_src;
            p_dst_by = p_dst + 4 * x + 4 * y * i_dst;

            i_best = -1;
            predict_4x4_mode_available( h->mb.i_neighbour, idx, predict_mode, &i_max );
            for( i = 0; i < i_max; i++ )
            {
                int i_sad;
                int i_mode;

                i_mode = predict_mode[i];

                /* we do the prediction */
                h->predict_4x4[i_mode]( p_dst_by, i_dst );

                /* we calculate diff and get the square sum of the diff */
                i_sad = h->pixf.satd[PIXEL_4x4]( p_dst_by, i_dst, p_src_by, i_src );

                i_sad += res->i_lambda * (i_pred_mode == x264_mb_pred_mode4x4_fix[i_mode] ? 1 : 4);

                /* if i_score is lower it is better */
                if( i_best == -1 || i_best > i_sad )
                {
                    res->i_predict4x4[x][y] = i_mode;
                    i_best = i_sad;
                }
            }
            res->i_sad_i4x4 += i_best;

            /* we need to encode this mb now (for next ones) */
            h->predict_4x4[res->i_predict4x4[x][y]]( p_dst_by, i_dst );
            x264_mb_encode_i4x4( h, idx, res->i_qp );

            /* we need to store the 'fixed' version */
            h->mb.cache.intra4x4_pred_mode[x264_scan8[idx]] =
                x264_mb_pred_mode4x4_fix[res->i_predict4x4[x][y]];
        }
        res->i_sad_i4x4 += res->i_lambda * 24;    /* from JVT (SATD0) */
    }
}

static void x264_mb_analyse_inter_p_p8x8( x264_t *h, x264_mb_analysis_t *res )
{
    const unsigned int flags = h->param.analyse.inter;
    const int i_ref = res->i_ref_p16x16;
    const int i_fref = h->mb.pic.i_fdec[0];
    uint8_t  *p_fref = h->mb.pic.p_fref[0][i_ref][0];
    const int i_img = h->mb.pic.i_img[0];
    uint8_t  *p_img = h->mb.pic.p_img[0];

    int i;

    res->i_ref_p8x8 = i_ref;
    res->i_sad_p8x8 = 0;

    for( i = 0; i < 4; i++ )
    {
        const int y = 8 * (i / 2);
        const int x = 8 * (i % 2);

        int satd[4] = { -1, -1, -1, -1 };
        int mv[4][4][2];
        int best;
        int idx;

        if( flags & X264_ANALYSE_P8x8 )
        {
            x264_mb_predict_mv( h, 0, 4*i, 2, mv[0][0] );
            satd[0] = h->me( h, &p_fref[y*i_fref+x], i_fref, &p_img[y*i_img+x], i_img,
                             PIXEL_8x8, res->i_lambda, &mv[0][0][0], &mv[0][0][1] );
            satd[0]+= res->i_lambda * bs_size_ue( 0 );
        }
        if( flags & X264_ANALYSE_P8x4 )
        {
            x264_mb_predict_mv( h, 0, 4*i+0, 2, mv[1][0] );
            satd[1] = h->me( h, &p_fref[y*i_fref+x], i_fref, &p_img[y*i_img+x], i_img,
                             PIXEL_8x4, res->i_lambda, &mv[1][0][0], &mv[1][0][1] );
            x264_macroblock_cache_mv( h, 2*(i%2), 2*(i/2), 2, 1, 0, mv[1][0][0], mv[1][0][1] );

            x264_mb_predict_mv( h, 0, 4*i+2, 2, mv[1][1] );
            satd[1]+= h->me( h, &p_fref[(y+4)*i_fref+x], i_fref, &p_img[(y+4)*i_img+x], i_img,
                             PIXEL_8x4, res->i_lambda, &mv[1][1][0], &mv[1][1][1] );
            satd[1]+= res->i_lambda * bs_size_ue( 1 );
        }
        if( flags & X264_ANALYSE_P4x8 )
        {
            x264_mb_predict_mv( h, 0, 4*i+0, 1, mv[2][0] );
            satd[2] = h->me( h, &p_fref[y*i_fref+x], i_fref, &p_img[y*i_img+x], i_img,
                             PIXEL_4x8, res->i_lambda, &mv[2][0][0], &mv[2][0][1] );
            x264_macroblock_cache_mv( h, 2*(i%2), 2*(i/2), 1, 2, 0, mv[2][0][0], mv[2][0][1] );

            x264_mb_predict_mv( h, 0, 4*i+1, 2, mv[2][1] );
            satd[2]+= h->me( h, &p_fref[y*i_fref+x+4], i_fref, &p_img[y*i_img+x+4], i_img,
                             PIXEL_4x8, res->i_lambda, &mv[2][1][0], &mv[2][1][1] );
            satd[2]+= res->i_lambda * bs_size_ue( 2 );
        }
        if( flags & X264_ANALYSE_P4x4 )
        {
            /* 4x4 mode */
            x264_mb_predict_mv( h, 0, 4*i+0, 1, mv[3][0] );
            satd[3] = h->me( h, &p_fref[y*i_fref+x], i_fref, &p_img[y*i_img+x], i_img,
                             PIXEL_4x4, res->i_lambda, &mv[3][0][0], &mv[3][0][1] );
            x264_macroblock_cache_mv( h, 2*(i%2)+0, 2*(i/2)+0, 1, 1, 0, mv[3][0][0], mv[3][0][1] );

            x264_mb_predict_mv( h, 0, 4*i+1, 1, mv[3][1] );
            satd[3]+= h->me( h, &p_fref[y*i_fref+x+4], i_fref, &p_img[y*i_img+x+4], i_img,
                             PIXEL_4x4, res->i_lambda, &mv[3][1][0], &mv[3][1][1] );
            x264_macroblock_cache_mv( h, 2*(i%2)+1, 2*(i/2)+0, 1, 1, 0, mv[3][1][0], mv[3][1][1] );

            x264_mb_predict_mv( h, 0, 4*i+2, 1, mv[3][2] );
            satd[3]+= h->me( h, &p_fref[(y+4)*i_fref+x], i_fref, &p_img[(y+4)*i_img+x], i_img,
                             PIXEL_4x4, res->i_lambda, &mv[3][2][0], &mv[3][2][1] );
            x264_macroblock_cache_mv( h, 2*(i%2)+0, 2*(i/2)+1, 1, 1, 0, mv[3][2][0], mv[3][2][1] );

            x264_mb_predict_mv( h, 0, 4*i+3, 1, mv[3][3] );
            satd[3]+= h->me( h, &p_fref[(y+4)*i_fref+x+4], i_fref, &p_img[(y+4)*i_img+x+4], i_img,
                             PIXEL_4x4, res->i_lambda, &mv[3][3][0], &mv[3][3][1] );
            satd[3]+= res->i_lambda * bs_size_ue( 3 );
        }

        /* Search best mode */
        for( best = 0; best < 4 && satd[best] == -1; ) best++;
        for( idx = best + 1; idx < 4; idx++ )
        {
            if( satd[idx] >= 0 && satd[idx] < satd[best] )
            {
                best = idx;
            }
        }

        res->i_sad_p8x8 += satd[best];

        switch( best )
        {
            case 0: /* 8x8 */
                res->i_sub_partition_p8x8[i] = D_L0_8x8;
                res->i_mv_p8x8[i][0][0] = mv[0][0][0]; res->i_mv_p8x8[i][0][1] = mv[0][0][1];
                x264_macroblock_cache_mv( h, 2*(i%2)+0, 2*(i/2)+0, 2, 2, 0, mv[0][0][0], mv[0][0][1] );
                break;
            case 1: /* 8x4 */
                res->i_sub_partition_p8x8[i] = D_L0_8x4;
                res->i_mv_p8x8[i][0][0] = mv[1][0][0]; res->i_mv_p8x8[i][0][1] = mv[1][0][1];
                res->i_mv_p8x8[i][1][0] = mv[1][1][0]; res->i_mv_p8x8[i][1][1] = mv[1][1][1];
                x264_macroblock_cache_mv( h, 2*(i%2)+0, 2*(i/2)+0, 2, 1, 0, mv[1][0][0], mv[1][0][1] );
                x264_macroblock_cache_mv( h, 2*(i%2)+0, 2*(i/2)+1, 2, 1, 0, mv[1][1][0], mv[1][1][1] );
                break;
            case 2: /* 4x8 */
                res->i_sub_partition_p8x8[i] = D_L0_4x8;
                res->i_mv_p8x8[i][0][0] = mv[2][0][0]; res->i_mv_p8x8[i][0][1] = mv[2][0][1];
                res->i_mv_p8x8[i][1][0] = mv[2][1][0]; res->i_mv_p8x8[i][1][1] = mv[2][1][1];
                x264_macroblock_cache_mv( h, 2*(i%2)+0, 2*(i/2)+0, 1, 2, 0, mv[2][0][0], mv[2][0][1] );
                x264_macroblock_cache_mv( h, 2*(i%2)+1, 2*(i/2)+0, 1, 2, 0, mv[2][1][0], mv[2][1][1] );
                break;
            case 3: /* 4x4 */
                res->i_sub_partition_p8x8[i] = D_L0_4x4;
                res->i_mv_p8x8[i][0][0] = mv[3][0][0]; res->i_mv_p8x8[i][0][1] = mv[3][0][1];
                res->i_mv_p8x8[i][1][0] = mv[3][1][0]; res->i_mv_p8x8[i][1][1] = mv[3][1][1];
                res->i_mv_p8x8[i][2][0] = mv[3][2][0]; res->i_mv_p8x8[i][2][1] = mv[3][2][1];
                res->i_mv_p8x8[i][3][0] = mv[3][3][0]; res->i_mv_p8x8[i][3][1] = mv[3][3][1];
                x264_macroblock_cache_mv( h, 2*(i%2)+0, 2*(i/2)+0, 1, 1, 0, mv[3][0][0], mv[3][0][1] );
                x264_macroblock_cache_mv( h, 2*(i%2)+1, 2*(i/2)+0, 1, 1, 0, mv[3][1][0], mv[3][1][1] );
                x264_macroblock_cache_mv( h, 2*(i%2)+0, 2*(i/2)+1, 1, 1, 0, mv[3][2][0], mv[3][2][1] );
                x264_macroblock_cache_mv( h, 2*(i%2)+1, 2*(i/2)+1, 1, 1, 0, mv[3][3][0], mv[3][3][1] );
                break;
        }
    }
    res->i_sad_p8x8 += 4*res->i_lambda * bs_size_te( h->sh.i_num_ref_idx_l0_active - 1, i_ref );
}

static void x264_mb_analyse_inter_p( x264_t *h, x264_mb_analysis_t *res )
{
    const unsigned int flags = h->param.analyse.inter;
    int i_ref;

    /* int res */
    res->i_sad_p16x16 = -1;
    res->i_sad_p16x8  = -1;
    res->i_sad_p8x16  = -1;
    res->i_sad_p8x8   = -1;

    /* 16x16 Search on all ref frame */
    h->mb.i_partition = D_16x16; /* predict mv */
    for( i_ref = 0; i_ref < h->i_ref0; i_ref++ )
    {
        int i_sad;
        int mvp[2];
        int mvx, mvy;

        /* Get the predicted MV */
        x264_macroblock_cache_ref( h, 0, 0, 4, 4, 0, i_ref );
        x264_mb_predict_mv( h, 0, 0, 4, mvp );

        mvx = mvp[0]; mvy = mvp[1];
        i_sad = h->me( h, h->mb.pic.p_fref[0][i_ref][0], h->mb.pic.i_fdec[0],
                          h->mb.pic.p_img[0],            h->mb.pic.i_img[0],
                          PIXEL_16x16, res->i_lambda, &mvx, &mvy );
        if( mvx == mvp[0] && mvy == mvp[1] )
        {
            i_sad -= 16 * res->i_lambda;
        }
        i_sad += res->i_lambda * bs_size_te( h->sh.i_num_ref_idx_l0_active - 1, i_ref );

        if( res->i_sad_p16x16 == -1 || i_sad < res->i_sad_p16x16 )
        {
            res->i_sad_p16x16   = i_sad;
            res->i_ref_p16x16   = i_ref;
            res->i_mv_p16x16[0] = mvx;
            res->i_mv_p16x16[1] = mvy;
        }
    }

    /* Now do the rafinement (using the ref found in 16x16 mode) */
    i_ref = res->i_ref_p16x16;
    x264_macroblock_cache_ref( h, 0, 0, 4, 4, 0, i_ref );

    /* try 16x8 */
    /* XXX we test i_predict16x16 to try shape with the same direction than edge
     * We should do a better algo of course (the one with edge dectection to be used
     * for intra mode too)
     * */

    if( (flags & X264_ANALYSE_P16x8) && ( !(flags & X264_ANALYSE_SMART_PSUB) || res->i_predict16x16 != I_PRED_16x16_V ) )
    {
        h->mb.i_partition = D_16x8;   /* predict mv */

        res->i_ref_p16x8   = i_ref;
        x264_mb_predict_mv( h, 0, 0, 4, res->i_mv_p16x8[0] );
        res->i_sad_p16x8 = h->me( h,
                                  h->mb.pic.p_fref[0][i_ref][0], h->mb.pic.i_fdec[0],
                                  h->mb.pic.p_img[0],            h->mb.pic.i_img[0],
                                  PIXEL_16x8,
                                  res->i_lambda,
                                  &res->i_mv_p16x8[0][0], &res->i_mv_p16x8[0][1] );
        /* needed for next predict */
        x264_macroblock_cache_mv( h, 0, 0, 4, 2, 0, res->i_mv_p16x8[0][0], res->i_mv_p16x8[0][1] );

        x264_mb_predict_mv( h, 0, 8, 4, res->i_mv_p16x8[1] );
        res->i_sad_p16x8+= h->me( h,
                                  &h->mb.pic.p_fref[0][i_ref][0][8*h->mb.pic.i_fdec[0]], h->mb.pic.i_fdec[0],
                                  &h->mb.pic.p_img[0][8*h->mb.pic.i_img[0]],    h->mb.pic.i_img[0],
                                  PIXEL_16x8,
                                  res->i_lambda,
                                  &res->i_mv_p16x8[1][0], &res->i_mv_p16x8[1][1] );

        res->i_sad_p16x8 += 2*res->i_lambda * bs_size_te( h->sh.i_num_ref_idx_l0_active - 1, i_ref );
    }

    /* try 8x16 */
    if( (flags & X264_ANALYSE_P8x16) && ( !(flags & X264_ANALYSE_SMART_PSUB) || res->i_predict16x16 != I_PRED_16x16_H ) )
    {
        h->mb.i_partition = D_8x16;   /* predict mv */

        res->i_ref_p8x16   = i_ref;
        x264_mb_predict_mv( h, 0, 0, 2, res->i_mv_p8x16[0] );
        res->i_sad_p8x16 = h->me( h,
                                  h->mb.pic.p_fref[0][i_ref][0], h->mb.pic.i_fdec[0],
                                  h->mb.pic.p_img[0],            h->mb.pic.i_img[0],
                                  PIXEL_8x16,
                                  res->i_lambda,
                                  &res->i_mv_p8x16[0][0], &res->i_mv_p8x16[0][1] );
        /* needed for next predict */
        x264_macroblock_cache_mv( h, 0, 0, 2, 4, 0, res->i_mv_p8x16[0][0], res->i_mv_p8x16[0][1] );

        x264_mb_predict_mv( h, 0, 4, 2, res->i_mv_p8x16[1] );
        res->i_sad_p8x16+= h->me( h,
                                  &h->mb.pic.p_fref[0][i_ref][0][8], h->mb.pic.i_fdec[0],
                                  &h->mb.pic.p_img[0][8],    h->mb.pic.i_img[0],
                                  PIXEL_8x16,
                                  res->i_lambda,
                                  &res->i_mv_p8x16[1][0], &res->i_mv_p8x16[1][1] );

        res->i_sad_p8x16 += 2*res->i_lambda * bs_size_te( h->sh.i_num_ref_idx_l0_active - 1, i_ref );
    }

    /* a bit heuristique : if 4x4 is prefered, the block is probably not homegenous
     * for now disabled because too slow for too few bits saved */
    if( ( flags & (X264_ANALYSE_P8x8|X264_ANALYSE_P8x4|X264_ANALYSE_P4x8|X264_ANALYSE_P4x4) ) &&
        ( !(flags & X264_ANALYSE_SMART_PSUB) || !(flags & X264_ANALYSE_I4x4 ) || res->i_sad_i4x4 < res->i_sad_i16x16 ) )
    {
        h->mb.i_partition = D_8x8;
        x264_mb_analyse_inter_p_p8x8( h,res );
    }
}

static void x264_mb_analyse_inter_b( x264_t *h, x264_mb_analysis_t *res )
{
    uint8_t pix1[16*16], pix2[16*16];

    int i_ref;
    /* int i_ref0, i_ref1; */

    int mvp[2];

    res->i_sad_b16x16_l0 = -1;
    res->i_sad_b16x16_l1 = -1;
    res->i_sad_b16x16_bi = -1;

    /* 16x16 L0 Search on all ref frame */
    h->mb.i_partition = D_16x16;
    for( i_ref = 0; i_ref < h->i_ref0; i_ref++ )
    {
        int i_sad;

        /* Get the predicted MV */
        x264_macroblock_cache_ref( h, 0, 0, 4, 4, 0, i_ref );
        x264_mb_predict_mv( h, 0, 0, 4, mvp );

        i_sad = h->me( h, h->mb.pic.p_fref[0][i_ref][0], h->mb.pic.i_fdec[0],
                          h->mb.pic.p_img[0],            h->mb.pic.i_img[0],
                          PIXEL_16x16, res->i_lambda, &mvp[0], &mvp[1] );
        i_sad += res->i_lambda * bs_size_te( h->sh.i_num_ref_idx_l0_active - 1, i_ref );

        if( res->i_sad_b16x16_l0 == -1 || i_sad < res->i_sad_b16x16_l0 )
        {
            res->i_sad_b16x16_l0   = i_sad;
            res->i_ref_b16x16_l0   = i_ref;
            res->i_mv_b16x16_l0[0] = mvp[0];
            res->i_mv_b16x16_l0[1] = mvp[1];
        }
    }

    /* 16x16 L1 Search on all ref frame */
    h->mb.i_partition = D_16x16;
    for( i_ref = 0; i_ref < h->i_ref1; i_ref++ )
    {
        int i_sad;

        /* Get the predicted MV */
        x264_macroblock_cache_ref( h, 0, 0, 4, 4, 1, i_ref );
        x264_mb_predict_mv( h, 1, 0, 4, mvp );

        i_sad = h->me( h, h->mb.pic.p_fref[1][i_ref][0], h->mb.pic.i_fdec[0],
                          h->mb.pic.p_img[0],            h->mb.pic.i_img[0],
                          PIXEL_16x16, res->i_lambda, &mvp[0], &mvp[1] );
        i_sad += res->i_lambda * bs_size_te( h->sh.i_num_ref_idx_l1_active - 1, i_ref );

        if( res->i_sad_b16x16_l1 == -1 || i_sad < res->i_sad_b16x16_l1 )
        {
            res->i_sad_b16x16_l1   = i_sad;
            res->i_ref_b16x16_l1   = i_ref;
            res->i_mv_b16x16_l1[0] = mvp[0];
            res->i_mv_b16x16_l1[1] = mvp[1];
        }
    }

    /* calculate i_sad_b16x16_bi */
    h->mc[MC_LUMA]( h->mb.pic.p_fref[0][res->i_ref_b16x16_l0][0], h->mb.pic.i_fdec[0],
                    pix1, 16,
                    res->i_mv_b16x16_l0[0],
                    res->i_mv_b16x16_l0[1],
                    16, 16 );
    h->mc[MC_LUMA]( h->mb.pic.p_fref[1][res->i_ref_b16x16_l0][0], h->mb.pic.i_fdec[0],
                    pix2, 16,
                    res->i_mv_b16x16_l1[0],
                    res->i_mv_b16x16_l1[1],
                    16, 16 );
    h->pixf.avg[PIXEL_16x16]( pix1, 16, pix2, 16 );

    res->i_sad_b16x16_bi = h->pixf.sad[PIXEL_16x16]( h->mb.pic.p_img[0], h->mb.pic.i_img[0], pix1, 16 );

    x264_mb_predict_mv( h, 0, 0, 4, mvp );
    res->i_sad_b16x16_bi += res->i_lambda * bs_size_te( h->sh.i_num_ref_idx_l0_active - 1, res->i_ref_b16x16_l0 );

    x264_mb_predict_mv( h, 1, 0, 4, mvp );
    res->i_sad_b16x16_bi += res->i_lambda * bs_size_te( h->sh.i_num_ref_idx_l1_active - 1, res->i_ref_b16x16_l1 );
}

/*****************************************************************************
 * x264_macroblock_analyse:
 *****************************************************************************/
void x264_macroblock_analyse( x264_t *h )
{
    x264_mb_analysis_t analysis;
    int i;

    /* qp TODO */
    h->mb.qp[h->mb.i_mb_xy] = x264_clip3( h->pps->i_pic_init_qp + h->sh.i_qp_delta + 0, 0, 51 );

    /* init analysis */
    analysis.i_qp = h->mb.qp[h->mb.i_mb_xy];
    analysis.i_lambda = i_qp0_cost_table[analysis.i_qp];

    /*--------------------------- Do the analysis ---------------------------*/
    x264_mb_analyse_intra( h, &analysis );
    if( h->sh.i_type == SLICE_TYPE_P )
    {
        x264_mb_analyse_inter_p( h, &analysis );
    }
    else if( h->sh.i_type == SLICE_TYPE_B )
    {
        x264_mb_analyse_inter_b( h, &analysis );
    }

    /*-------------------- Chose the macroblock mode ------------------------*/
#define BEST_TYPE( type, partition, satd ) \
        if( satd != -1 && satd < i_satd ) \
        {   \
            i_satd = satd;  \
            h->mb.i_type = type; \
            h->mb.i_partition = partition; \
        }
    if( h->sh.i_type == SLICE_TYPE_I )
    {
        if( analysis.i_sad_i4x4 >= 0 &&  analysis.i_sad_i4x4 < analysis.i_sad_i16x16 )
            h->mb.i_type = I_4x4;
        else
            h->mb.i_type = I_16x16;
    }
    else if( h->sh.i_type == SLICE_TYPE_P )
    {
        int i_satd = analysis.i_sad_i16x16;
        h->mb.i_type = I_16x16;

        BEST_TYPE( I_4x4, -1,      analysis.i_sad_i4x4 );
        BEST_TYPE( P_L0,  D_16x16, analysis.i_sad_p16x16 );
        BEST_TYPE( P_L0,  D_16x8 , analysis.i_sad_p16x8  );
        BEST_TYPE( P_L0,  D_8x16 , analysis.i_sad_p8x16  );
        BEST_TYPE( P_8x8, D_8x8  , analysis.i_sad_p8x8   );
    }
    else    /* B */
    {
        int i_satd = analysis.i_sad_i16x16;
        h->mb.i_type = I_16x16;

        BEST_TYPE( I_4x4,   -1,      analysis.i_sad_i4x4 );
        BEST_TYPE( B_L0_L0, D_16x16, analysis.i_sad_b16x16_l0 );
        BEST_TYPE( B_L1_L1, D_16x16, analysis.i_sad_b16x16_l1 );
        BEST_TYPE( B_BI_BI, D_16x16, analysis.i_sad_b16x16_bi );
    }
#undef BEST_TYPE

    if( IS_INTRA( h->mb.i_type ) )
    {
        x264_mb_analyse_intra_chroma( h, &analysis );
    }

    /*-------------------- Update MB from the analysis ----------------------*/
    h->mb.type[h->mb.i_mb_xy] = h->mb.i_type;
    switch( h->mb.i_type )
    {
        case I_4x4:
            for( i = 0; i < 16; i++ )
            {
                h->mb.cache.intra4x4_pred_mode[x264_scan8[i]] =
                    analysis.i_predict4x4[block_idx_x[i]][block_idx_y[i]];
            }
            h->mb.i_chroma_pred_mode = analysis.i_predict8x8;
            break;
        case I_16x16:
            h->mb.i_intra16x16_pred_mode = analysis.i_predict16x16;
            h->mb.i_chroma_pred_mode = analysis.i_predict8x8;
            break;
        case P_L0:
            switch( h->mb.i_partition )
            {
                case D_16x16:
                    x264_macroblock_cache_ref( h, 0, 0, 4, 4, 0, analysis.i_ref_p16x16 );
                    x264_macroblock_cache_mv ( h, 0, 0, 4, 4, 0, analysis.i_mv_p16x16[0], analysis.i_mv_p16x16[1] );
                    break;
                case D_16x8:
                    x264_macroblock_cache_ref( h, 0, 0, 4, 2, 0, analysis.i_ref_p16x8);
                    x264_macroblock_cache_mv ( h, 0, 0, 4, 2, 0, analysis.i_mv_p16x8[0][0], analysis.i_mv_p16x8[0][1] );

                    x264_macroblock_cache_ref( h, 0, 2, 4, 2, 0, analysis.i_ref_p16x8);
                    x264_macroblock_cache_mv ( h, 0, 2, 4, 2, 0, analysis.i_mv_p16x8[1][0], analysis.i_mv_p16x8[1][1] );
                    break;
                case D_8x16:
                    x264_macroblock_cache_ref( h, 0, 0, 2, 4, 0, analysis.i_ref_p8x16 );
                    x264_macroblock_cache_mv ( h, 0, 0, 2, 4, 0, analysis.i_mv_p8x16[0][0], analysis.i_mv_p8x16[0][1] );

                    x264_macroblock_cache_ref( h, 2, 0, 2, 4, 0, analysis.i_ref_p8x16 );
                    x264_macroblock_cache_mv ( h, 2, 0, 2, 4, 0, analysis.i_mv_p8x16[1][0], analysis.i_mv_p8x16[1][1] );
                    break;
                default:
                    fprintf( stderr, "internal error\n" );
                    break;
            }
            break;

        case P_8x8:
            for( i = 0; i < 4; i++ )
            {
                const int x = 2*(i%2);
                const int y = 2*(i/2);

                h->mb.i_sub_partition[i] = analysis.i_sub_partition_p8x8[i];

                x264_macroblock_cache_ref( h, x, y, 2, 2, 0, analysis.i_ref_p8x8 );
                switch( analysis.i_sub_partition_p8x8[i] )
                {
                    case D_L0_8x8:
                        x264_macroblock_cache_mv( h, x, y, 2, 2, 0, analysis.i_mv_p8x8[i][0][0], analysis.i_mv_p8x8[i][0][1] );
                        break;
                    case D_L0_8x4:
                        x264_macroblock_cache_mv( h, x, y+0, 2, 1, 0, analysis.i_mv_p8x8[i][0][0], analysis.i_mv_p8x8[i][0][1] );
                        x264_macroblock_cache_mv( h, x, y+1, 2, 1, 0, analysis.i_mv_p8x8[i][1][0], analysis.i_mv_p8x8[i][1][1] );
                        break;
                    case D_L0_4x8:
                        x264_macroblock_cache_mv( h, x+0, y, 1, 2, 0, analysis.i_mv_p8x8[i][0][0], analysis.i_mv_p8x8[i][0][1] );
                        x264_macroblock_cache_mv( h, x+1, y, 1, 2, 0, analysis.i_mv_p8x8[i][1][0], analysis.i_mv_p8x8[i][1][1] );
                        break;
                    case D_L0_4x4:
                        x264_macroblock_cache_mv( h, x+0, y+0, 1, 1, 0, analysis.i_mv_p8x8[i][0][0], analysis.i_mv_p8x8[i][0][1] );
                        x264_macroblock_cache_mv( h, x+1, y+0, 1, 1, 0, analysis.i_mv_p8x8[i][1][0], analysis.i_mv_p8x8[i][1][1] );
                        x264_macroblock_cache_mv( h, x+0, y+1, 1, 1, 0, analysis.i_mv_p8x8[i][2][0], analysis.i_mv_p8x8[i][2][1] );
                        x264_macroblock_cache_mv( h, x+1, y+1, 1, 1, 0, analysis.i_mv_p8x8[i][3][0], analysis.i_mv_p8x8[i][3][1] );
                        break;
                    default:
                        fprintf( stderr, "internal error\n" );
                        break;
                }
            }
            break;
        case B_L0_L0:
            switch( h->mb.i_partition )
            {
                case D_16x16:
                    x264_macroblock_cache_ref( h, 0, 0, 4, 4, 0, analysis.i_ref_b16x16_l0 );
                    x264_macroblock_cache_mv ( h, 0, 0, 4, 4, 0, analysis.i_mv_b16x16_l0[0], analysis.i_mv_b16x16_l0[1] );

                    x264_macroblock_cache_ref( h, 0, 0, 4, 4, 1, -1 );
                    x264_macroblock_cache_mv ( h, 0, 0, 4, 4, 1,  0, 0 );
                    break;
                default:
                    fprintf( stderr, "internal error\n" );
                    break;
            }
            break;
        case B_L1_L1:
            switch( h->mb.i_partition )
            {
                case D_16x16:
                    x264_macroblock_cache_ref( h, 0, 0, 4, 4, 0, -1 );
                    x264_macroblock_cache_mv ( h, 0, 0, 4, 4, 0,  0, 0 );

                    x264_macroblock_cache_ref( h, 0, 0, 4, 4, 1, analysis.i_ref_b16x16_l1 );
                    x264_macroblock_cache_mv ( h, 0, 0, 4, 4, 1, analysis.i_mv_b16x16_l1[0], analysis.i_mv_b16x16_l1[1] );
                    break;

                default:
                    fprintf( stderr, "internal error\n" );
                    break;
            }
            break;
        case B_BI_BI:
            switch( h->mb.i_partition )
            {
                case D_16x16:
                    x264_macroblock_cache_ref( h, 0, 0, 4, 4, 0, analysis.i_ref_b16x16_l0 );
                    x264_macroblock_cache_mv ( h, 0, 0, 4, 4, 0, analysis.i_mv_b16x16_l0[0], analysis.i_mv_b16x16_l0[1] );

                    x264_macroblock_cache_ref( h, 0, 0, 4, 4, 1, analysis.i_ref_b16x16_l1 );
                    x264_macroblock_cache_mv ( h, 0, 0, 4, 4, 1, analysis.i_mv_b16x16_l1[0], analysis.i_mv_b16x16_l1[1] );
                    break;

                default:
                    fprintf( stderr, "internal error\n" );
                    break;
            }
            break;

        default:
            fprintf( stderr, "internal error\n" );
            break;
    }
}

